/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema, Romain Lenglet
 */

package chat;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManager;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

import chat.gui.DisplayMessage;
import chat.gui.PublishMessage;

/**
 * Chat component implementation
 */
public class ChatClientImpl extends AbstractComponent
    implements
      Push,
      PublishMessage,
      ChatAttributeController
{

  private ChunkFactoryReference<ChatChunk> chatChunkFactory;

  // ---------------------------------------------------------------------------
  // Attribute fields
  // ---------------------------------------------------------------------------

  String                                   userName;

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  Push                                     outPushItf;
  MessageManager                           messageManagerItf;
  DisplayMessage                           displayMessageItf;

  // ---------------------------------------------------------------------------
  // Implementation of the Push interface
  // ---------------------------------------------------------------------------

  /**
   * @see Push#push(Message)
   */
  public void push(Message message) throws PushException
  {
    ChatChunk chatChunk = (ChatChunk) messageManagerItf.getChunk(message,
        ChatChunk.DEFAULT_NAME);
    if (chatChunk != null)
    {
      displayMessageItf.displayMessage(chatChunk.getFrom(), chatChunk
          .getMessage());
    }
    messageManagerItf.deleteMessage(message);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the PublishMessage interface
  // ---------------------------------------------------------------------------

  /**
   * @see PublishMessage#publishMessage(String)
   */
  public void publishMessage(String message)
  {
    Message msg = messageManagerItf.createMessage();
    if (chatChunkFactory == null)
    {
      chatChunkFactory = messageManagerItf.getChunkFactory(ChatChunk.class);
    }
    ChatChunk chatChunk = messageManagerItf.createChunk(chatChunkFactory);
    chatChunk.set(userName, message);
    messageManagerItf.addChunk(msg, ChatChunk.DEFAULT_NAME, chatChunk);
    try
    {
      outPushItf.push(msg);
    }
    catch (PushException e)
    {
      logger.log(BasicLevel.ERROR, e);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the ChatAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see ChatAttributeController#getUserName()
   */
  public String getUserName()
  {
    return userName;
  }

  /**
   * @see ChatAttributeController#setUserName(String)
   */
  public void setUserName(String name)
  {
    this.userName = name;
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(Push.OUT_PUSH_ITF_NAME))
    {
      outPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
    else if (clientItfName.equals(DisplayMessage.ITF_NAME))
    {
      displayMessageItf = (DisplayMessage) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{Push.OUT_PUSH_ITF_NAME, MessageManager.ITF_NAME,
        DisplayMessage.ITF_NAME};
  }

}
