/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.message;

import java.util.Iterator;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.util.EmptyIterator;
import org.objectweb.dream.util.EmptyStringArray;
import org.objectweb.dream.util.IteratorNoRemove;

/**
 * Basic message manager implementation.
 */
public class MessageManagerImpl extends AbstractComponent
    implements
      MessageManager,
      MessageManagerAttributeController
{
  private static final Iterator<Message> EMPTY_SUB_MESSAGE_ITERATOR = new EmptyIterator<Message>();
  final ChunkFactoryRepository           chunkFactoryRepository     = new ChunkFactoryRepository();

  private boolean                        check                      = false;

  // ---------------------------------------------------------------------------
  // Implementation of the MessageManager interface
  // ---------------------------------------------------------------------------

  // Message creation ----------------------------------------------------------

  /**
   * @see MessageManager#createMessage()
   */
  public Message createMessage()
  {
    final BasicMessageImpl message = new BasicMessageImpl(this);
    message.incrementReferenceCounter();
    return message;
  }

  /**
   * @see MessageManager#deleteMessage(Message)
   */
  public void deleteMessage(final Message message)
  {
    checkOwner(message);
    if (message.decrementReferenceCounter())
    {
      final Iterator<String> names = message.getNameIterator();
      while (names.hasNext())
      {
        final String chunkName = names.next();
        final AbstractChunk chunk = getChunk(message, chunkName);
        names.remove();
        deleteChunk(chunk);
      }
      final Iterator<Message> subMessages = message.getSubMessageIterator();
      while (subMessages.hasNext())
      {
        deleteMessage(subMessages.next());
      }
      recycleMessageInstance(message);
    }
  }

  /**
   * @see MessageManager#duplicateMessage(Message, boolean)
   */
  @SuppressWarnings("unchecked")
  public Message duplicateMessage(final Message message, final boolean clone)
  {
    checkOwner(message);
    if (clone)
    {
      final Message newMessage = createMessage();
      final Iterator<String> names = message.getNameIterator();
      while (names.hasNext())
      {
        final String chunkName = names.next();
        final AbstractChunk chunk = getChunk(message, chunkName);
        final AbstractChunk newChunk = chunk.getFactory().createChunk();
        chunk.transfertStateTo(newChunk);
        newMessage.addChunk(chunkName, newChunk);
      }
      final Iterator<Message> subMessages = message.getSubMessageIterator();
      while (subMessages.hasNext())
      {
        newMessage.addSubMessage(duplicateMessage(subMessages.next(), true));
      }
      // TODO Auto-generated method stub
      return null;
    }
    message.incrementReferenceCounter();
    return message;
  }

  // Chunk creation ------------------------------------------------------------

  /**
   * @see MessageManager#getChunkFactory(Class)
   */
  public synchronized <T extends AbstractChunk<T>> ChunkFactoryReference<T> getChunkFactory(
      final Class<T> chunkType)
  {
    ChunkFactoryReference<T> factory = chunkFactoryRepository
        .getFactory(chunkType);
    if (factory == null)
    {
      factory = createChunkFactory(chunkType);
      chunkFactoryRepository.addFactory(chunkType, factory);
    }
    return factory;
  }

  /**
   * @see MessageManager#createChunk(ChunkFactoryReference)
   */
  public <T extends AbstractChunk<T>> T createChunk(
      final ChunkFactoryReference<T> chunkFactory)
  {
    return chunkFactory.createChunk();
  }

  /**
   * @see MessageManager#cloneChunk(AbstractChunk)
   */
  public <T extends AbstractChunk<T>> T cloneChunk(final T chunk)
  {
    checkOwner(chunk);
    final T clone = chunk.getFactory().createChunk();
    chunk.transfertStateTo(clone);
    return clone;
  }

  /**
   * @see MessageManager#deleteChunk(AbstractChunk)
   */
  public void deleteChunk(final AbstractChunk< ? > chunk)
  {
    checkOwner(chunk);
    chunk.recycleChunk();
  }

  // Chunk manipulation --------------------------------------------------------

  /**
   * @see MessageManager#isEmpty(Message)
   */
  public boolean isEmpty(final Message message)
  {
    checkOwner(message);
    return message.isEmpty();
  }

  /**
   * @see MessageManager#getChunk(Message, String)
   */
  public AbstractChunk getChunk(final Message message, final String chunkName)
  {
    checkOwner(message);
    checkNull(chunkName);
    final AbstractChunk chunk = message.getChunk(chunkName);
    if (chunk == null)
    {
      return null;
    }
    return chunk;
  }

  /**
   * @see MessageManager#addChunk(Message, String, AbstractChunk)
   */
  public void addChunk(final Message message, final String chunkName,
      final AbstractChunk chunk)
  {
    checkOwner(message);
    checkOwner(chunk);
    checkNull(chunkName);
    if (message.getChunk(chunkName) != null)
    {
      throw new ChunkAlreadyExistException(chunkName);
    }
    message.addChunk(chunkName, chunk);
  }

  /**
   * @see MessageManager#addOrReplaceChunk(Message, String, AbstractChunk)
   */
  public AbstractChunk addOrReplaceChunk(final Message message,
      final String chunkName, final AbstractChunk chunk)
  {
    checkOwner(message);
    checkOwner(chunk);
    checkNull(chunkName);
    final AbstractChunk previous = message.getChunk(chunkName);
    message.addChunk(chunkName, chunk);
    return previous;
  }

  /**
   * @see MessageManager#removeChunk(Message, String)
   */
  public AbstractChunk removeChunk(final Message message, final String chunkName)
  {
    checkOwner(message);
    checkNull(chunkName);
    final AbstractChunk chunk = message.getChunk(chunkName);
    if (chunk == null)
    {
      throw new UnknownChunkException(chunkName);
    }
    message.removeChunk(chunkName);
    return chunk;
  }

  /**
   * @see MessageManager#getChunkNameIterator(Message)
   */
  public Iterator<String> getChunkNameIterator(final Message message)
  {
    checkOwner(message);
    checkOwner(message);
    final Iterator<String> iter = message.getNameIterator();
    if (!iter.hasNext())
    {
      return EmptyIterator.EMPTY_STRING_ITERATOR;
    }
    return new IteratorNoRemove<String>(iter);
  }

  // Sub message manipulation --------------------------------------------------

  /**
   * @see MessageManager#addSubMessage(Message, Message)
   */
  public void addSubMessage(final Message message, final Message subMessage)
  {
    checkOwner(message);
    checkOwner(subMessage);
    checkSubMessage(message, subMessage);
    message.addSubMessage(subMessage);
  }

  /**
   * @see MessageManager#getSubMessageIterator(Message)
   */
  public Iterator<Message> getSubMessageIterator(final Message message)
  {
    checkOwner(message);
    final Iterator<Message> subMessageIterator = message.getSubMessageIterator();
    if (!subMessageIterator.hasNext())
    {
      return EMPTY_SUB_MESSAGE_ITERATOR;
    }
    return new IteratorNoRemove<Message>(subMessageIterator);
  }

  /**
   * @see MessageManager#removeSubMessage(Message, Message)
   */
  public void removeSubMessage(final Message message, final Message subMessage)
  {
    checkOwner(message);
    checkOwner(subMessage);
    final boolean removed = message.removeSubMessage(subMessage);
    if (!removed)
    {
      throw new IllegalArgumentException(
          "Given subMessage is not a sub message of the given message");
    }
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  protected final void checkOwner(final Message message)
  {
    checkNull(message);
    if (message.getOwner() != this)
    {
      throw new InvalidReference(
          "Message is not owned by this message manager", null);
    }
  }

  protected final void checkOwner(final AbstractChunk< ? > chunk)
  {
    checkNull(chunk);
    if (chunk.getOwner() != this)
    {
      throw new InvalidReference("Chunk is not owned by this message manager",
          null);
    }
  }

  protected final void checkNull(final String chunkName)
  {
    if (chunkName == null)
    {
      throw new NullPointerException("Given chunk name is null");
    }
  }

  protected final void checkNull(final AbstractChunk chunk)
  {
    if (chunk == null)
    {
      throw new NullPointerException("Given chunk is null");
    }
  }

  protected final void checkNull(final Message message)
  {
    if (message == null)
    {
      throw new NullPointerException("Given message reference is null");
    }
  }

  protected final void checkSubMessage(final Message message,
      final Message subMessage)
  {
    if (!check) return;
    if (message == subMessage)
    {
      throw new IllegalArgumentException("message and subMessage are equals");
    }
    if (subMessage.isSubMessage(message))
    {
      throw new IllegalArgumentException(
          "message is a sub message of subMessage");
    }
    final Iterator<Message> iterator = subMessage.getSubMessageIterator();
    while (iterator.hasNext())
    {
      checkSubMessage(message, iterator.next());
    }
  }

  protected <T extends AbstractChunk<T>> ChunkFactoryReference<T> createChunkFactory(
      final Class<T> chunkType)
  {
    try
    {
      return new BasicChunkFactoryImpl<T>(chunkType, this);
    }
    catch (Exception e)
    {
      throw new IllegalArgumentException(e);
    }
  }

  protected void recycleMessageInstance(final Message message)
  {
  }

  // ---------------------------------------------------------------------------
  // Implementation of the AttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see MessageManagerAttributeController#setDoCheck(boolean)
   */
  public void setDoCheck(final boolean doCheck)
  {
    this.check = doCheck;
  }

  /**
   * @see MessageManagerAttributeController#getDoCheck()
   */
  public boolean getDoCheck()
  {
    return check;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return EmptyStringArray.EMPTY_STRING_ARRAY;
  }
}
