/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s):
 */

package org.objectweb.dream.protocol.bus.logicalClock;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManager;
import org.objectweb.dream.util.EmptyStringArray;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Basic Implementation of a matrix clock. Stamps used by this component are
 * integer arrays. This implementation is not efficient if the number of process
 * involved in the causality order is important.
 */
public class MatrixClock extends AbstractComponent implements LogicalClock
{

  protected int                                      matrix[][];
  protected int                                      myAddress;
  protected int                                      size;
  protected String                                   logicalClockChunkName = LogicalClockChunk.DEFAULT_NAME;
  protected MessageManager                           messageManager;
  protected ChunkFactoryReference<LogicalClockChunk> logicalClockChunkFactory;

  /**
   * @param size the bus size;
   * @param myAddress the local address in the bus.
   * @param logicalClockChunkName the name of the chunk to add in the message.
   * @param messageManager the message manager.
   */
  public MatrixClock(int size, int myAddress, String logicalClockChunkName,
      MessageManager messageManager)
  {
    this.size = size;
    this.myAddress = myAddress;
    this.logicalClockChunkName = logicalClockChunkName;
    this.messageManager = messageManager;
    matrix = new int[size][size];
  }

  // ---------------------------------------------------------------------------
  // Implementation of LogicalClock interface
  // ---------------------------------------------------------------------------

  /**
   * @see LogicalClock#testRecvMessage(Message)
   */
  public synchronized int testRecvMessage(Message message)
  {
    LogicalClockChunk chunk = (LogicalClockChunk) messageManager.getChunk(
        message, logicalClockChunkName);
    int todo = testRecvMatrix(chunk.getLogicalClock(), chunk.getFrom());
    if (todo == DELIVER)
    {
      messageManager.removeChunk(message, logicalClockChunkName);
      messageManager.deleteChunk(chunk);
    }
    return todo;
  }

  /**
   * @see LogicalClock#stamp(Message, int)
   */
  public synchronized void stamp(Message message, int to)
  {
    if (logicalClockChunkFactory == null)
    {
      logicalClockChunkFactory = messageManager
          .getChunkFactory(LogicalClockChunk.class);
    }
    LogicalClockChunk chunk = messageManager
        .createChunk(logicalClockChunkFactory);
    chunk.setFrom(myAddress);
    chunk.setLogicalClock(getStamp(to));
    messageManager.addChunk(message, logicalClockChunkName, chunk);
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return EmptyStringArray.EMPTY_STRING_ARRAY;
  }

  // ---------------------------------------------------------------------------
  // Overriding of LifeCycleController methods
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
   */
  public void startFc() throws IllegalLifeCycleException
  {
    if (matrix == null)
    {
      throw new IllegalLifeCycleException("Matrix size has not been set");
    }
    super.startFc();
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  protected int testRecvMatrix(Object stamp, int from)
  {
    int[][] stampTab = (int[][]) stamp;
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "testRecvMatrix from=" + from + " to="
          + myAddress + " stamp=");
      logMatrix(stampTab);
      logger.log(BasicLevel.DEBUG, "matrix state :");
      logMatrix(matrix);
    }
    if (matrix[from][myAddress] + 1 < stampTab[from][myAddress])
    {
      // There is other messages from the same node to deliver before this one.
      logger.log(BasicLevel.DEBUG, "return WAIT_TO_DELIVER");
      return WAIT_TO_DELIVER;
    }
    else if (matrix[from][myAddress] + 1 == stampTab[from][myAddress])
    {
      // Verify that all messages to be delivered to this node and known by this
      // message have already been delivered.
      for (int k = 0; k < matrix.length; k++)
      {
        if ((k != from) && (stampTab[k][myAddress] > matrix[k][myAddress]))
        {
          logger.log(BasicLevel.DEBUG, "return WAIT_TO_DELIVER");
          return WAIT_TO_DELIVER;
        }
      }
    }
    else
    {
      logger.log(BasicLevel.DEBUG, "return ALREADY_DELIVERED");
      return ALREADY_DELIVERED;
    }
    // the message is ready to be delivered, so updates the matrix clock
    for (int i = 0; i < stampTab.length; i++)
    {
      for (int j = 0; j < stampTab.length; j++)
      {
        if (stampTab[i][j] > matrix[i][j])
        {
          matrix[i][j] = stampTab[i][j];
        }
      }
    }
    logger.log(BasicLevel.DEBUG, "return DELIVER");
    return DELIVER;
  }

  protected Object getStamp(int to)
  {
    // increment matrix clock
    matrix[myAddress][to]++;
    // copy the matrix and set the copy as stamp.
    int[][] stamp = new int[matrix.length][matrix.length];
    for (int i = 0; i < stamp.length; i++)
    {
      System.arraycopy(matrix[i], 0, stamp[i], 0, matrix.length);
    }
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "getStamp from=" + myAddress + " to=" + to
          + " returned stamp=");
      logMatrix(stamp);
    }
    return stamp;
  }

  private void logMatrix(final int[][] matrix)
  {
    if (!logger.isLoggable(BasicLevel.DEBUG))
    {
      return;
    }
    for (int i = 0; i < matrix.length; i++)
    {
      StringBuffer stringBuffer = new StringBuffer("[");
      for (int j = 0; j < matrix.length; j++)
      {
        stringBuffer.append(matrix[i][j]);
        if (j < matrix.length - 1)
        {
          stringBuffer.append(", ");
        }
      }
      stringBuffer.append("]");
      logger.log(BasicLevel.DEBUG, stringBuffer);
    }
  }

}
