/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.bus.logicalClock;

import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManager;
import org.objectweb.dream.protocol.BindException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.protocol.bus.BusOutgoingPush;
import org.objectweb.dream.protocol.bus.BusProtocol;
import org.objectweb.dream.protocol.channel.ChannelProtocol;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Generic logical clock protocol implementation.
 */
public class ProtocolImpl extends AbstractComponent
    implements
      LogicalClockBusProtocol
{

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  protected BusProtocol    lowerLevelBusProtocolItf;
  protected MessageManager messageManagerItf;

  // ---------------------------------------------------------------------------
  // Implementation of the BusProtocol interface
  // ---------------------------------------------------------------------------

  /**
   * @see BusProtocol#bind(ExportIdentifier, IncomingPush, Map)
   */
  public BusOutgoingPush bind(ExportIdentifier exportId,
      IncomingPush toClientPush, Map<String, Object> hints)
      throws InvalidExportIdentifierException, BindException
  {
    if (!(exportId instanceof LogicalClockExportIdentifier))
    {
      throw new InvalidExportIdentifierException(
          "Must be a BusExportIdentifier");
    }
    Session session = new Session(toClientPush);
    BusOutgoingPush lowerPush = lowerLevelBusProtocolItf.bind(exportId
        .getNextExportIds()[0], session, hints);

    // create logical clock
    int clockType = ((LogicalClockExportIdentifier) exportId).getClockType();
    String chunkName = ((LogicalClockExportIdentifier) exportId).getChunkName();
    LogicalClock logicalClock;
    switch (clockType)
    {
      case LogicalClockExportIdentifier.FIFO_CLOCK_TYPE :
        logicalClock = new FIFOClock(lowerPush.getBusSize(), lowerPush
            .getLocalAddress(), chunkName, messageManagerItf);
        break;

      case LogicalClockExportIdentifier.MATRIX_CAUSAL_CLOCK_TYPE :
        logicalClock = new MatrixClock(lowerPush.getBusSize(), lowerPush
            .getLocalAddress(), chunkName, messageManagerItf);
        break;
      default :
        throw new InvalidExportIdentifierException("Unknown clock type : "
            + clockType);
    }

    // initialize session
    session.initialize(logicalClock, lowerPush);

    return session;
  }

  /**
   * @see BusProtocol#createExportIdentifier(Map, ExportIdentifier[])
   */
  public ExportIdentifier createExportIdentifier(Map<String, Object> info,
      ExportIdentifier[] next) throws InvalidExportIdentifierException
  {
    int clockType;
    String chunkName = LogicalClockChunk.DEFAULT_NAME;

    Object o = info.get(CLOCK_TYPE_HINT_KEY);
    if (o == null)
    {
      throw new InvalidExportIdentifierException(
          "Unable to find clock type in info map");
    }
    if (o instanceof String)
    {
      String typeName = (String) o;
      if (typeName
          .equalsIgnoreCase(LogicalClockExportIdentifier.FIFO_CLOCK_TYPE_NAME))
      {
        clockType = LogicalClockExportIdentifier.FIFO_CLOCK_TYPE;
      }
      else if (typeName
          .equalsIgnoreCase(LogicalClockExportIdentifier.MATRIX_CAUSAL_CLOCK_TYPE_NAME))
      {
        clockType = LogicalClockExportIdentifier.MATRIX_CAUSAL_CLOCK_TYPE;
      }
      else
      {
        throw new InvalidExportIdentifierException("Unknown clock type name : "
            + typeName);
      }
    }
    else if (o instanceof Number)
    {
      clockType = ((Number) o).intValue();
    }
    else
    {
      throw new InvalidExportIdentifierException(
          "Invalid clock type in info map, must be either a Number or a String");
    }

    o = info.get(LOGICAL_CLOCK_CHUNK_NAME_HINT_TYPE);
    if (o != null)
    {
      if (!(o instanceof String))
      {
        throw new InvalidExportIdentifierException(
            "Invalid logical clock chunk name in info map, must be a String");
      }
      chunkName = (String) o;
    }

    if (next == null)
    {
      throw new InvalidExportIdentifierException(
          "Invalid array of next identifier.");
    }
    ExportIdentifier lowerId = lowerLevelBusProtocolItf.createExportIdentifier(
        info, next);
    return createExportIdentifier(clockType, chunkName, lowerId);
  }

  /**
   * @see LogicalClockBusProtocol#createExportIdentifier(int, String,
   *      ExportIdentifier)
   */
  public LogicalClockExportIdentifier createExportIdentifier(int clockType,
      String chunkName, ExportIdentifier lowerId)
  {
    return new LogicalClockExportIdentifier(clockType, chunkName, lowerId);
  }

  // ---------------------------------------------------------------------------
  // Inner classes
  // ---------------------------------------------------------------------------

  protected class Session implements BusOutgoingPush, IncomingPush
  {
    protected IncomingPush    upperIncomingPush;

    protected boolean         initialized      = false;
    protected BusOutgoingPush lowerBusOutgoingPush;
    protected LogicalClock    logicalClock;

    protected List<Message>   waitingToDeliver = new LinkedList<Message>();

    protected Session(IncomingPush upperIncomingPush)
    {
      this.upperIncomingPush = upperIncomingPush;
    }

    protected synchronized void initialize(LogicalClock logicalClock,
        BusOutgoingPush lowerBusOutgoingPush)
    {
      this.logicalClock = logicalClock;
      this.lowerBusOutgoingPush = lowerBusOutgoingPush;
      initialized = true;
      notifyAll();
    }

    // -------------------------------------------------------------------------
    // Implementation of the BusOutgoingPush interface
    // -------------------------------------------------------------------------

    /**
     * @see BusOutgoingPush#outgoingPush(Message, int)
     */
    public void outgoingPush(Message message, int to) throws PushException
    {
      logicalClock.stamp(message, to);
      lowerBusOutgoingPush.outgoingPush(message, to);
    }

    /**
     * @see BusOutgoingPush#outgoingClose(IncomingPush)
     */
    public synchronized void outgoingClose(IncomingPush incomingPush)
        throws IOException
    {
      lowerBusOutgoingPush.outgoingClose(this);
    }

    /**
     * @see BusOutgoingPush#getLocalAddress()
     */
    public int getLocalAddress()
    {
      return lowerBusOutgoingPush.getLocalAddress();
    }

    /**
     * @see BusOutgoingPush#getBusSize()
     */
    public int getBusSize()
    {
      return lowerBusOutgoingPush.getBusSize();
    }

    // -------------------------------------------------------------------------
    // Implementation of the IncomingPush interface
    // -------------------------------------------------------------------------

    /**
     * @see IncomingPush#incomingPush(Message)
     */
    public synchronized void incomingPush(Message message) throws PushException
    {
      while (!initialized)
      {
        try
        {
          wait();
        }
        catch (InterruptedException e)
        {
          throw new PushException(
              "Interrupted while waiting for the session to be initialized", e);
        }
      }
      int todo = logicalClock.testRecvMessage(message);
      if (todo == LogicalClock.DELIVER)
      {
        upperIncomingPush.incomingPush(message);

        // scan waiting to deliver message if they can be delivered.
        boolean loop = true;
        while (loop)
        {
          loop = false;
          Iterator iter = waitingToDeliver.iterator();
          while (iter.hasNext())
          {
            Message msg = (Message) iter.next();
            if (logger.isLoggable(BasicLevel.DEBUG))
            {
              logger.log(BasicLevel.DEBUG, "Try to deliver message : " + msg);
            }
            todo = logicalClock.testRecvMessage(msg);
            if (todo == LogicalClock.DELIVER)
            {
              iter.remove();
              upperIncomingPush.incomingPush(msg);
              // Matrix clock has changed, we have to rescan the list.
              loop = true;
              break;
            }
            else if (todo == LogicalClock.ALREADY_DELIVERED)
            {
              Error.bug(logger);
            }
            logger.log(BasicLevel.DEBUG, "Message can't be delivered yet");
          }
        }
      }
      else if (todo == LogicalClock.WAIT_TO_DELIVER)
      {
        // insert in the waiting list
        waitingToDeliver.add(message);
        if (logger.isLoggable(BasicLevel.DEBUG))
        {
          logger.log(BasicLevel.DEBUG, "block message " + message);
        }
      }
      else
      {
        // message already delivered.
        if (logger.isLoggable(BasicLevel.DEBUG))
        {
          logger.log(BasicLevel.DEBUG, "message already delivered " + message);
        }
        messageManagerItf.deleteMessage(message);
      }

    }

    /**
     * @see IncomingPush#incomingClosed(Object, Exception)
     */
    public synchronized void incomingClosed(Object outgoingPush,
        Exception exception)
    {
      // never called on bus protocol
      Error.bug(logger);
    }

  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{MessageManager.ITF_NAME,
        ChannelProtocol.LOWER_PROTOCOL_ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
    else if (clientItfName.equals(ChannelProtocol.LOWER_PROTOCOL_ITF_NAME))
    {
      lowerLevelBusProtocolItf = (BusProtocol) serverItf;
    }
    super.bindFc(clientItfName, serverItf);
  }

}
