/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.channel;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.control.activity.Util;
import org.objectweb.dream.control.activity.manager.ThreadPoolManager;
import org.objectweb.dream.control.activity.task.AbstractTask;
import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.thread.ThreadPoolOverflowException;
import org.objectweb.dream.protocol.IPExportIdentifier;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Implementation of {@link ConnectionFactory}.
 */
public class ConnectionFactoryImpl extends AbstractComponent
    implements
      ConnectionFactory
{

  private static final boolean                      DEFENSIVE_CHECK  = true;
  protected SessionFactory                          firstAvailable   = null;
  protected SessionFactory                          lastAvailable    = null;
  protected Map<IPExportIdentifier, SessionFactory> sessionFactories = new HashMap<IPExportIdentifier, SessionFactory>();

  protected ThreadPoolManager                       threadPoolManager;
  protected ListenTask                              listenTask       = new ListenTask();

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  protected ConnectionManager                       connectionManagerItf;

  // ---------------------------------------------------------------------------
  // Implementation of the ConnectionFactory interface
  // ---------------------------------------------------------------------------

  /**
   * @see ConnectionFactory#addServerSocket(ServerSocket, IPExportIdentifier)
   */
  public void addServerSocket(ServerSocket serverSocket,
      IPExportIdentifier identifier)
  {
    SessionFactory sessionFactory = new SessionFactory(serverSocket, identifier);
    sessionFactories.put(identifier, sessionFactory);
    setAvailable(sessionFactory);
    if (threadPoolManager == null)
    {
      Error.bug(logger);
    }
    try
    {
      logger.log(BasicLevel.DEBUG, "add a thread in thread pool");
      threadPoolManager.addThread(listenTask);
    }
    catch (ThreadPoolOverflowException e)
    {
      logger.log(BasicLevel.WARN, "Unable to add listen thread", e);
    }
    catch (IllegalTaskException e)
    {
      Error.bug(logger, e);
    }
  }

  /**
   * @see ConnectionFactory#removeServerSocket(IPExportIdentifier)
   */
  public synchronized void removeServerSocket(IPExportIdentifier identifier)
      throws InvalidExportIdentifierException
  {
    SessionFactory sessionFactory = sessionFactories.remove(identifier);
    if (sessionFactory == null)
    {
      throw new InvalidExportIdentifierException("Unknown export identifier.");
    }
    try
    {
      sessionFactory.closeSession();
    }
    catch (IOException e)
    {
      logger.log(BasicLevel.WARN,
          "exception caught while closing server socket", e);
    }

    // remove session factory from available list
    SessionFactory previous = null;
    SessionFactory available = firstAvailable;
    while (available != null)
    {
      if (available == sessionFactory)
      {
        if (previous == null)
        {
          firstAvailable = sessionFactory.nextAvailable;
        }
        else
        {
          previous.nextAvailable = sessionFactory.nextAvailable;
        }
        if (sessionFactory.nextAvailable == null)
        {
          lastAvailable = previous;
        }
        break;
      }
      previous = available;
      available = available.nextAvailable;
    }
  }

  // ---------------------------------------------------------------------------
  // Inner Class
  // ---------------------------------------------------------------------------

  protected class SessionFactory
  {
    private boolean            closed = false;
    private SessionFactory     nextAvailable;
    private ServerSocket       serverSocket;
    private IPExportIdentifier exportIdentifier;

    protected SessionFactory(ServerSocket socket, IPExportIdentifier identifier)
    {
      serverSocket = socket;
      exportIdentifier = identifier;
    }

    protected Socket listen() throws IOException
    {
      return serverSocket.accept();
    }

    synchronized void closeSession() throws IOException
    {
      closed = true;
      serverSocket.close();
    }
  }

  protected class ListenTask extends AbstractTask
  {

    public ListenTask()
    {
      super("TCP/IP listen task");
    }

    // -------------------------------------------------------------------------
    // Implementation of the Task interface
    // -------------------------------------------------------------------------

    /**
     * @see org.objectweb.dream.control.activity.task.Task#execute(Object)
     */
    public Object execute(Object hints) throws InterruptedException
    {
      SessionFactory sessionFactory = getAvailable();
      if (sessionFactory == null)
      {
        return STOP_EXECUTING;
      }
      Socket socket;
      try
      {
        socket = sessionFactory.listen();
      }
      catch (IOException e)
      {
        if (!sessionFactory.closed)
        {
          try
          {
            sessionFactory.serverSocket.close();
          }
          catch (IOException e1)
          {
            // ignore
          }
          logger.log(BasicLevel.ERROR, e);
        }

        return STOP_EXECUTING;
      }
      connectionManagerItf.newConnection(socket,
          sessionFactory.exportIdentifier);
      setAvailable(sessionFactory);
      return EXECUTE_AGAIN;
    }

    /**
     * @see org.objectweb.dream.control.activity.task.Task#registered(Object)
     */
    public void registered(Object controlItf)
    {
      threadPoolManager = (ThreadPoolManager) controlItf;
    }

  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  protected synchronized void setAvailable(SessionFactory session)
  {
    if (firstAvailable == null)
    {
      if (DEFENSIVE_CHECK && lastAvailable != null)
      {
        Error.bug(logger);
      }
      firstAvailable = session;
      lastAvailable = session;
    }
    else
    {
      if ((DEFENSIVE_CHECK)
          && (session.nextAvailable != null || lastAvailable.nextAvailable != null))
      {
        Error.bug(logger);
      }
      lastAvailable.nextAvailable = session;
      lastAvailable = session;
    }
  }

  protected synchronized SessionFactory getAvailable()
  {
    if (firstAvailable == null)
    {
      if (DEFENSIVE_CHECK && lastAvailable != null)
      {
        Error.bug(logger);
      }
      return null;
    }
    SessionFactory available = firstAvailable;
    firstAvailable = available.nextAvailable;
    if (firstAvailable == null)
    {
      lastAvailable = null;
    }
    available.nextAvailable = null;
    return available;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{ConnectionManager.ITF_NAME};
  }

  /**
   * @see AbstractComponent#bindFc(String, Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(ConnectionManager.ITF_NAME))
    {
      connectionManagerItf = (ConnectionManager) serverItf;
    }
  }

  // -------------------------------------------------------------------------
  // Overridden methods
  // -------------------------------------------------------------------------

  /**
   * @see AbstractComponent#beforeFirstStart(Component)
   */
  protected void beforeFirstStart(Component componentItf)
      throws IllegalLifeCycleException
  {
    try
    {
      Map<String, Object> hints = new HashMap<String, Object>();
      hints.put("thread", "pool");
      Util.addTask(componentItf, listenTask, hints);
    }
    catch (Exception e)
    {
      throw new IllegalLifeCycleException("Can't add task");
    }
  }

  /**
   * @see org.objectweb.dream.control.lifecycle.PrepareStopLifeCycleController#prepareStopFc()
   */
  public synchronized void prepareStopFc() throws IllegalLifeCycleException
  {
    logger.log(BasicLevel.DEBUG, "Going to stop, close every opened sockets");
    Iterator iter = sessionFactories.values().iterator();
    while (iter.hasNext())
    {
      try
      {
        ((SessionFactory) iter.next()).closeSession();
      }
      catch (IOException e)
      {
        // ignore
      }
    }
  }

}
