/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.channel;

import org.objectweb.dream.protocol.BindException;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.IPExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.protocol.OutgoingPush;

/**
 * Defines constants used as hints key that can be given to a TCP/IP protocol.
 */
public interface TCPIPProtocol extends ChannelProtocol
{
  /**
   * This integer hint, used at export time, can specify the listening port on
   * which the TCP/IP protocol will wait incoming connection for an exported
   * channel. Can also be used with the
   * {@link org.objectweb.dream.protocol.Protocol#createExportIdentifier} method
   * to specify the port of the export identifier to create.
   */
  String PORT             = "port";

  /**
   * This integer hint, used at export time, can specify the range of listening
   * ports on which the TCP/IP protocol will try to open server socket. <br>
   * If this hints is not specified and a port has been specified (with
   * {@link #PORT} key), only the specified port will be tried before aborting.
   * <br>
   * if neither this hint nor the port hint are specified, the TCP/IP protocol
   * will try any port starting from its current listening port.
   */
  String PORT_RANGE       = "range";

  /**
   * Key name for value used by the
   * {@link org.objectweb.dream.protocol.Protocol#createExportIdentifier} method
   * to specify the address of the export identifier to create. The value
   * associated with this key can be an {@link java.net.InetAddress} or a
   * {@link String}.
   */
  String ADDRESS          = "address";

  /**
   * This integer hint, used at bind time, can specify the number of connection
   * retries before aborting.
   */
  String CONNECTION_RETRY = "connection-retry";

  /**
   * This boolean hint, used either at export or bind time, can specify if the
   * Nagle's algorithm is disabled.
   * 
   * @see java.net.Socket#setTcpNoDelay(boolean)
   */
  String TCP_NO_DELAY     = "tcp-no-delay";

  /**
   * This integer hint, used either at export or bind time, can specify the
   * <code>SO_TIMEOUT</code> parameter of created sockets.
   * 
   * @see java.net.Socket#setSoTimeout(int)
   */
  String SO_TIMEOUT       = "so-timeout";

  /**
   * This integer hint, used either at export or bind time, can specify the
   * <code>SO_LINGER</code> parameter of created sockets. A negative value
   * disable the linger on close.
   * 
   * @see java.net.Socket#setSoLinger(boolean, int)
   */
  String SO_LINGER        = "so-linger";

  /**
   * Utility method. Exports the channel through the given port.
   * 
   * @param channel the channel to export
   * @param port the listening port on which the TCP/IP protocol will wait
   *          incoming connection for an exported channel. <br>
   *          If null, the current listening port of the protocol is used.
   * @param range the range of listening ports on which the TCP/IP protocol will
   *          try to open server socket. <br>
   *          If null, only one port will be tried before aborting. If negative,
   *          range is <i>infinite</i>.
   * @return An export identifier.
   * @throws ExportException if the export process fails.
   * @see ChannelProtocol#export(ChannelFactory, java.util.Map)
   */
  IPExportIdentifier export(ChannelFactory channel, int port, int range)
      throws ExportException;

  /**
   * Utility method.
   * 
   * @see ChannelProtocol#bind(org.objectweb.dream.protocol.ExportIdentifier,
   *      IncomingPush, java.util.Map)
   */
  OutgoingPush bind(String hostName, int port, IncomingPush toClientPush)
      throws InvalidExportIdentifierException, BindException;

  /**
   * Utility method.
   * 
   * @see ChannelProtocol#createExportIdentifier(java.util.Map,
   *      org.objectweb.dream.protocol.ExportIdentifier[])
   */
  IPExportIdentifier createExportIdentifier(String hostName, int port);
}
