/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.channel;

import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.IOPushException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.control.activity.Util;
import org.objectweb.dream.control.activity.manager.ThreadPoolManager;
import org.objectweb.dream.control.activity.task.AbstractTask;
import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.thread.ThreadPoolOverflowException;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManager;
import org.objectweb.dream.message.codec.CodecInputOutput;
import org.objectweb.dream.message.codec.CodecInputOutputImpl;
import org.objectweb.dream.message.codec.MessageCodec;
import org.objectweb.dream.protocol.BindException;
import org.objectweb.dream.protocol.ExceptionChunk;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.IPExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.protocol.OutgoingPush;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Implementation of the TCP/IP protocol.
 */
public class TCPIPProtocolImpl extends AbstractComponent
    implements
      TCPIPProtocolAttributeController,
      TCPIPProtocol,
      ConnectionManager
{
  protected static final String                     BIND_REPLY_CHUNK_NAME = "tcpip-bind-reply";
  private static final boolean                      DEFENSIVE_CHECK       = true;

  protected static String                           hostName;

  static
  {
    try
    {
      hostName = InetAddress.getLocalHost().getHostName();
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }

  protected ChunkFactoryReference<ExceptionChunk>   exceptionChunkFactory;
  protected Set<Session>                            sessions              = new HashSet<Session>();
  protected Session                                 firstAvailable        = null;
  protected Session                                 lastAvailable         = null;

  protected ThreadPoolManager                       threadPoolManager;
  protected ReaderTask                              readerTask            = new ReaderTask();
  protected Map<IPExportIdentifier, ChannelFactory> exportedChanels       = new HashMap<IPExportIdentifier, ChannelFactory>();

  // ---------------------------------------------------------------------------
  // Attribute fields
  // ---------------------------------------------------------------------------

  protected int                                     port;
  protected int                                     connectionRetry;
  protected boolean                                 tcpNoDelay;
  protected int                                     soTimeout;
  protected int                                     soLinger;

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  protected ConnectionFactory                       connectionFactoryItf;
  protected MessageManager                          messageManagerItf;
  protected MessageCodec                            messageCodecItf;

  // ---------------------------------------------------------------------------
  // Implementation of the TCPIPProtocol interface
  // ---------------------------------------------------------------------------

  /**
   * @see ChannelProtocol#export(ChannelFactory, Map)
   */
  public ExportIdentifier export(ChannelFactory channel,
      Map<String, Object> hints) throws ExportException
  {
    Number portN = null;
    Number rangeN = null;
    int port = 0;
    int range = 0;
    if (hints != null)
    {
      portN = (Number) hints.get(PORT);
      rangeN = (Number) hints.get(PORT_RANGE);
    }

    if (portN == null)
    {
      if (rangeN == null)
      {
        range = -1;
      }
      else
      {
        range = rangeN.intValue();
      }
    }
    else
    {
      port = portN.intValue();
      if (rangeN != null)
      {
        range = rangeN.intValue();
      }
    }
    return export(channel, port, range);
  }

  /**
   * @see ChannelProtocol#unexport(ExportIdentifier)
   */
  public void unexport(ExportIdentifier exportIdentifier)
      throws InvalidExportIdentifierException
  {
    if (!(exportIdentifier instanceof IPExportIdentifier))
    {
      throw new InvalidExportIdentifierException(
          "this protocol requires a IPExportIdentifier.", exportIdentifier);
    }
    connectionFactoryItf
        .removeServerSocket((IPExportIdentifier) exportIdentifier);
  }

  /**
   * @see ChannelProtocol#bind(ExportIdentifier, IncomingPush, Map)
   */
  public OutgoingPush bind(ExportIdentifier exportId,
      IncomingPush toClientPush, Map<String, Object> hints)
      throws InvalidExportIdentifierException, BindException
  {
    // TODO check local binding.
    // TODO use hints.
    if (!(exportId instanceof IPExportIdentifier))
    {
      throw new InvalidExportIdentifierException(
          "this protocol requires a IPExportIdentifier.", exportId);
    }
    IPExportIdentifier identifier = (IPExportIdentifier) exportId;
    Socket socket;
    try
    {
      socket = createConnection(identifier);
    }
    catch (IOException e)
    {
      throw new BindException(
          "Unable to open connection to specified export : ", exportId, e);
    }
    Session session;
    try
    {
      session = new Session(socket, identifier);
    }
    catch (IOException e)
    {
      silentClose(socket);
      throw new BindException("unable to create session for opened socket : ",
          exportId, e);
    }

    // wait for acknowledgment message.
    Message message;
    try
    {
      message = messageCodecItf.decode(session.codecInputOutput);
    }
    catch (IOException e)
    {
      silentClose(session);
      throw new BindException("Unable to receive acknowledgment message",
          exportId, e);
    }
    ExceptionChunk chunk = (ExceptionChunk) messageManagerItf.getChunk(message,
        BIND_REPLY_CHUNK_NAME);
    if (chunk == null)
    {
      silentClose(session);
      Error.error("First received message on socket does not contain "
          + "acknowledgment chunk", logger);
    }
    if (chunk.getException() != null)
    {
      throw new BindException("Remote exception", exportId, chunk
          .getException());
    }
    messageManagerItf.deleteMessage(message);
    initSession(session, toClientPush);
    return session;
  }

  /**
   * @see org.objectweb.dream.protocol.Protocol#createExportIdentifier(Map,
   *      ExportIdentifier[])
   */
  public ExportIdentifier createExportIdentifier(Map<String, Object> info,
      ExportIdentifier[] next) throws InvalidExportIdentifierException
  {
    if (next != null && next.length != 0)
    {
      throw new InvalidExportIdentifierException("TCP/IP protocol is a leaf "
          + "in the protocol graph, it cannot have next export identifier");
    }
    Object o;
    if (info == null || (o = info.get(ADDRESS)) == null)
    {
      throw new InvalidExportIdentifierException(
          "Can't find address in info map");
    }
    String hostName;
    if (o instanceof String)
    {
      hostName = (String) o;
    }
    else if (o instanceof InetAddress)
    {
      hostName = ((InetAddress) o).getCanonicalHostName();
    }
    else
    {
      throw new InvalidExportIdentifierException(
          "Invalid address object in info map, must be an InetAddress or"
              + " a String");
    }

    o = info.get(PORT);
    if (o == null)
    {
      throw new InvalidExportIdentifierException("Can't find port in info map");

    }
    if (!(o instanceof Number))
    {
      throw new InvalidExportIdentifierException(
          "Invalid port object in info map, must be a Number");
    }
    int port = ((Number) o).intValue();
    return createExportIdentifier(hostName, port);
  }

  /**
   * @see TCPIPProtocol#export(ChannelFactory, int, int)
   */
  public IPExportIdentifier export(ChannelFactory channel, int port, int range)
      throws ExportException
  {
    synchronized (this)
    {
      if (port == 0)
      {
        port = this.port;
        this.port++;
      }
    }
    int maxPort;
    if (range < 0)
    {
      maxPort = 65536;
    }
    else
    {
      maxPort = port + range;
    }
    ServerSocket serverSocket = null;
    for (int p = port; p <= maxPort; p++)
    {
      try
      {
        serverSocket = new ServerSocket(p);
        break;
      }
      catch (IOException e)
      {
        if (p == maxPort)
        {
          throw new ExportException("Unable to open server socket : "
              + e.getMessage());
        }
        logger.log(BasicLevel.INFO, "Unable to open server socket on port" + p,
            e);

      }
    }
    IPExportIdentifier exportIdentifier = new IPExportIdentifier(hostName,
        serverSocket.getLocalPort());
    exportedChanels.put(exportIdentifier, channel);
    connectionFactoryItf.addServerSocket(serverSocket, exportIdentifier);
    return exportIdentifier;
  }

  /**
   * @see TCPIPProtocol#bind(String, int, IncomingPush)
   */
  public OutgoingPush bind(String hostName, int port, IncomingPush toClientPush)
      throws InvalidExportIdentifierException, BindException
  {
    return bind(createExportIdentifier(hostName, port), toClientPush, null);
  }

  /**
   * @see TCPIPProtocol#createExportIdentifier(String, int)
   */
  public IPExportIdentifier createExportIdentifier(String hostName, int port)
  {
    return new IPExportIdentifier(hostName, port);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the ConnectionManager interface
  // ---------------------------------------------------------------------------

  /**
   * @see ConnectionManager#newConnection(Socket, IPExportIdentifier)
   */
  public void newConnection(Socket socket, IPExportIdentifier identifier)
  {
    ChannelFactory channel = exportedChanels.get(identifier);
    if (DEFENSIVE_CHECK && channel == null)
    {
      Error.bug(logger);
    }

    Message message = messageManagerItf.createMessage();
    ExceptionChunk chunk = newExceptionChunk();
    messageManagerItf.addChunk(message, BIND_REPLY_CHUNK_NAME, chunk);
    Session session;
    try
    {
      session = new Session(socket, identifier);
    }
    catch (IOException e)
    {
      silentClose(socket);
      logger.log(BasicLevel.ERROR,
          "Unable to create session for new incoming socket", e);
      return;
    }
    IncomingPush incomingPush = null;
    try
    {
      incomingPush = channel.instantiate(session);
    }
    catch (IOException e)
    {
      chunk.setException(e);
      logger.log(BasicLevel.DEBUG,
          "Exception caught while instantiating channel factory", e);
    }
    try
    {
      session.outgoingPush(message);
    }
    catch (IOPushException e)
    {
      logger.log(BasicLevel.ERROR,
          "Unable to send reply to client, close socket", e);
      silentClose(session);
      return;
    }
    if (incomingPush != null)
    {
      initSession(session, incomingPush);
    }
  }

  // ---------------------------------------------------------------------------
  // Utility method
  // ---------------------------------------------------------------------------

  private ExceptionChunk newExceptionChunk()
  {
    if (exceptionChunkFactory == null)
    {
      exceptionChunkFactory = messageManagerItf
          .getChunkFactory(ExceptionChunk.class);
    }
    return messageManagerItf.createChunk(exceptionChunkFactory);
  }

  // ---------------------------------------------------------------------------
  // Inner Class
  // ---------------------------------------------------------------------------

  protected class Session implements OutgoingPush
  {

    Session            nextAvailable;
    boolean            closed           = false;
    Socket             socket;
    CodecInputOutput   codecInputOutput = new CodecInputOutputImpl();
    IPExportIdentifier exportIdentifier;
    IncomingPush       upperIncomingPushItf;

    Session(Socket socket, IPExportIdentifier exportIdentifier)
        throws IOException
    {
      setSocket(socket, exportIdentifier);
    }

    void setSocket(Socket socket, IPExportIdentifier exportIdentifier)
        throws IOException
    {
      this.socket = socket;
      this.exportIdentifier = exportIdentifier;
      codecInputOutput.setInput(socket.getInputStream());
      codecInputOutput.setOutput(socket.getOutputStream());
    }

    synchronized void closeSession() throws IOException
    {
      closed = true;
      socket.close();
    }

    // -------------------------------------------------------------------------
    // Implementation of the OutgoingPush interface
    // -------------------------------------------------------------------------

    /**
     * @see OutgoingPush#outgoingPush(Message)
     */
    public synchronized void outgoingPush(Message message)
        throws IOPushException
    {
      if (closed)
      {
        throw new IOPushException("Session closed");
      }
      try
      {
        messageCodecItf.encode(codecInputOutput, message);
      }
      catch (IOException e)
      {
        if (!closed)
        {
          sessionError(this, e);
        }
        throw new IOPushException("unable to send message", e);
      }
      messageManagerItf.deleteMessage(message);
    }

    /**
     * @see OutgoingPush#outgoingClose(IncomingPush)
     */
    public void outgoingClose(IncomingPush incomingPush) throws IOException
    {
      closeSession();
    }

  }

  protected class ReaderTask extends AbstractTask
  {

    public ReaderTask()
    {
      super("TCP/IP reader task");
    }

    // -------------------------------------------------------------------------
    // Implementation of the Task interface
    // -------------------------------------------------------------------------

    /**
     * @see org.objectweb.dream.control.activity.task.Task#execute(Object)
     */
    public Object execute(Object hints) throws InterruptedException
    {
      Session session;
      synchronized (sessions)
      {
        session = getAvailable();
      }
      if (session == null)
      {
        logger.log(BasicLevel.ERROR, "no available session, stop executing");
        return STOP_EXECUTING;
      }
      if (session.closed)
      {
        return STOP_EXECUTING;
      }
      Message message;
      try
      {
        message = messageCodecItf.decode(session.codecInputOutput);
      }
      catch (IOException e)
      {
        synchronized (this)
        {
          if (!session.closed)
          {
            sessionError(session, e);
          }
        }
        return STOP_EXECUTING;
      }
      try
      {
        session.upperIncomingPushItf.incomingPush(message);
      }
      catch (PushException e)
      {
        logger.log(BasicLevel.ERROR,
            "Exception catch while pushing incoming message. delete message.",
            e);
        messageManagerItf.deleteMessage(message);
      }
      synchronized (sessions)
      {
        setAvailable(session);
      }
      return EXECUTE_AGAIN;
    }

    /**
     * @see org.objectweb.dream.control.activity.task.Task#registered(Object)
     */
    public void registered(Object controlItf)
    {
      setThreadPoolManager((ThreadPoolManager) controlItf);
    }
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  protected final void silentClose(Session session)
  {
    try
    {
      session.closeSession();
    }
    catch (IOException e)
    {
      // ignore
    }
  }

  protected final void silentClose(Socket socket)
  {
    try
    {
      socket.close();
    }
    catch (IOException e)
    {
      // ignore
    }
  }

  protected void setAvailable(Session session)
  {
    if (firstAvailable == null)
    {
      if (DEFENSIVE_CHECK && lastAvailable != null)
      {
        Error.bug(logger);
      }
      firstAvailable = session;
      lastAvailable = session;
    }
    else
    {
      if ((DEFENSIVE_CHECK)
          && (session.nextAvailable != null || lastAvailable.nextAvailable != null))
      {
        Error.bug(logger);
      }
      lastAvailable.nextAvailable = session;
      lastAvailable = session;
    }
  }

  protected Session getAvailable()
  {
    if (firstAvailable == null)
    {
      if (DEFENSIVE_CHECK && lastAvailable != null)
      {
        Error.bug(logger);
      }
      return null;
    }
    Session available = firstAvailable;
    firstAvailable = available.nextAvailable;
    if (firstAvailable == null)
    {
      lastAvailable = null;
    }
    available.nextAvailable = null;
    return available;
  }

  protected synchronized Session initSession(Session session,
      IncomingPush incomingPush)
  {
    session.upperIncomingPushItf = incomingPush;
    synchronized (sessions)
    {
      sessions.add(session);
      if (session.upperIncomingPushItf != null)
      {
        setAvailable(session);
        if (threadPoolManager != null)
        {
          try
          {
            logger.log(BasicLevel.DEBUG, "add a thread in thread pool");
            threadPoolManager.addThread(readerTask);
          }
          catch (Exception e)
          {
            Error.bug(logger, e);
          }
        }
      }
    }
    return session;
  }

  protected void setThreadPoolManager(ThreadPoolManager threadPoolManager)
  {
    this.threadPoolManager = threadPoolManager;
    synchronized (sessions)
    {
      for (int i = 0; i < sessions.size(); i++)
      {
        try
        {
          logger.log(BasicLevel.DEBUG, "add a thread in thread pool");
          threadPoolManager.addThread(readerTask);
        }
        catch (ThreadPoolOverflowException e)
        {
          logger.log(BasicLevel.WARN, "Unable to add reader thread", e);
        }
        catch (IllegalTaskException e)
        {
          Error.bug(logger, e);
        }
      }
    }
  }

  protected void sessionError(Session session, Exception exception)
  {
    logger.log(BasicLevel.INFO,
        "An error occurs while reading or writing message on TCP/IP socket",
        exception);
    synchronized (sessions)
    {
      boolean removed = sessions.remove(session);
      if (DEFENSIVE_CHECK && !removed)
      {
        Error.bug(logger);
      }
    }
    // session is a server session, inform channel.
    silentClose(session);
    if (session.upperIncomingPushItf != null)
    {
      session.upperIncomingPushItf.incomingClosed(session, exception);
    }
  }

  protected Socket createConnection(IPExportIdentifier exportIdentifier)
      throws IOException
  {
    for (int i = 0;; i++)
    {
      try
      {
        if (logger.isLoggable(BasicLevel.DEBUG))
        {
          logger.log(BasicLevel.DEBUG, "try to connect");
        }
        Socket sock = new Socket(exportIdentifier.getHostName(),
            exportIdentifier.getPort());
        setSocketOption(sock);
        if (logger.isLoggable(BasicLevel.DEBUG))
        {
          logger.log(BasicLevel.DEBUG, "connected to "
              + exportIdentifier.getHostName() + ":"
              + exportIdentifier.getPort());
        }
        return sock;
      }
      catch (IOException exc)
      {
        if (i >= connectionRetry)
        {
          throw exc;
        }
        logger.log(BasicLevel.INFO, "connection failed on address "
            + exportIdentifier.getHostName() + ":" + exportIdentifier.getPort()
            + " will retry");
        try
        {
          Thread.sleep(i * 250);
        }
        catch (InterruptedException ignored)
        {
          // ignored
        }
      }
    }
  }

  protected void setSocketOption(Socket sock) throws IOException
  {
    sock.setTcpNoDelay(tcpNoDelay);
    sock.setSoTimeout(soTimeout);
    if (soLinger >= 0)
    {
      sock.setSoLinger(true, soLinger);
    }
    else
    {
      sock.setSoLinger(false, 0);
    }
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "Options for socket ("
          + sock.getInetAddress() + "," + sock.getPort() + ") : tcpNoDelay = "
          + sock.getTcpNoDelay() + " soLinger = " + sock.getSoLinger()
          + " soTimeout = " + sock.getSoTimeout() + " sendBufferSize = "
          + sock.getSendBufferSize());
    }
  }

  // -------------------------------------------------------------------------
  // Overridden methods
  // -------------------------------------------------------------------------

  /**
   * @see AbstractComponent#beforeFirstStart(Component)
   */
  protected void beforeFirstStart(Component componentItf)
      throws IllegalLifeCycleException
  {
    try
    {
      Map<String, Object> hints = new HashMap<String, Object>();
      hints.put("thread", "pool");
      Util.addTask(componentItf, readerTask, hints);
    }
    catch (Exception e)
    {
      throw new IllegalLifeCycleException("Can't add task");
    }
  }

  /**
   * @see org.objectweb.dream.control.lifecycle.PrepareStopLifeCycleController#prepareStopFc()
   */
  public void prepareStopFc() throws IllegalLifeCycleException
  {
    synchronized (sessions)
    {
      logger.log(BasicLevel.DEBUG, "Going to stop, close every opened sockets");
      Iterator iter = sessions.iterator();
      while (iter.hasNext())
      {
        silentClose(((Session) iter.next()));
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the TCPIPProtocolAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see TCPIPProtocolAttributeController#getPort()
   */
  public int getPort()
  {
    return port;
  }

  /**
   * @see TCPIPProtocolAttributeController#setPort(int)
   */
  public void setPort(int port)
  {
    this.port = port;
  }

  /**
   * @see TCPIPProtocolAttributeController#getConnectionRetry()
   */
  public int getConnectionRetry()
  {
    return connectionRetry;
  }

  /**
   * @see TCPIPProtocolAttributeController#setConnectionRetry(int)
   */
  public void setConnectionRetry(int connectionRetry)
  {
    this.connectionRetry = connectionRetry;
  }

  /**
   * @see TCPIPProtocolAttributeController#getSoLinger()
   */
  public int getSoLinger()
  {
    return soLinger;
  }

  /**
   * @see TCPIPProtocolAttributeController#setSoLinger(int)
   */
  public void setSoLinger(int timeout)
  {
    soLinger = timeout;
  }

  /**
   * @see TCPIPProtocolAttributeController#getSoTimeout()
   */
  public int getSoTimeout()
  {
    return soTimeout;
  }

  /**
   * @see TCPIPProtocolAttributeController#setSoTimeout(int)
   */
  public void setSoTimeout(int timeout)
  {
    soTimeout = timeout;
  }

  /**
   * @see TCPIPProtocolAttributeController#getTcpNoDelay()
   */
  public boolean getTcpNoDelay()
  {
    return tcpNoDelay;
  }

  /**
   * @see TCPIPProtocolAttributeController#setTcpNoDelay(boolean)
   */
  public void setTcpNoDelay(boolean tcpNoDelay)
  {
    this.tcpNoDelay = tcpNoDelay;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{ConnectionFactory.ITF_NAME, MessageCodec.ITF_NAME,
        MessageManager.ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    if (clientItfName.equals(ConnectionFactory.ITF_NAME))
    {
      connectionFactoryItf = (ConnectionFactory) serverItf;
    }
    else if (clientItfName.equals(MessageCodec.ITF_NAME))
    {
      messageCodecItf = (MessageCodec) serverItf;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
    super.bindFc(clientItfName, serverItf);
  }
}
