/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.messagePassing;

import java.io.IOException;

import org.objectweb.dream.IOPushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;

/**
 * Specialized {@link org.objectweb.dream.protocol.OutgoingPush} interface for
 * message passing protocol. Since this <I>outgoing push</I> interface is not
 * connected to a given remote <i>incoming push</i> interface, a destination
 * identifier is associated to each outgoing message.
 */
public interface MessagePassingOutgoingPush
{
  /**
   * Returns the {@link ExportIdentifier} identifying the local access point.
   * <br>
   * More precisely, on remote host, messages sent using
   * {@link #outgoingPush outgoingPush} method with the export identifier
   * returned by this method, will be received by the local access point and
   * will be passed to the {@link IncomingPush} interface given at the creation
   * of the access point (see
   * {@link MessagePassingProtocol#export(IncomingPush, java.util.Map)}). <br>
   * May return <code>null</code> if the {@link IncomingPush} interface given
   * at the creation of the access point was <code>null</code> (depending on
   * protocol implementation).
   * 
   * @return an identifier of the local access point.
   */
  ExportIdentifier getLocalExportIdentifier();

  /**
   * Sends the given message to the access point identified by <code>to</code>.
   * 
   * @param message the message to send.
   * @param to the identifier of the access point that should received the
   *          message.
   * @throws InvalidExportIdentifierException if the destination is invalid.
   * @throws IOPushException if something goes wrong.
   */
  void outgoingPush(Message message, ExportIdentifier to)
      throws InvalidExportIdentifierException, IOPushException;

  /**
   * Closes the access point. The implementation of this method must not call
   * the {@link IncomingPush#incomingClosed} method, since this latter is only
   * used to inform a upper session that a lower session has been closed
   * unexpectedly.
   * 
   * @param incomingPush the interface on which incoming messages of the closed
   *          access point were received.
   * @throws IOException the closing process failed.
   */
  void outgoingClose(IncomingPush incomingPush) throws IOException;
}
