/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.pushwithreturn;

import java.util.Iterator;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.InterruptedPushException;
import org.objectweb.dream.Push;
import org.objectweb.dream.Push1;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManager;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This component has one {@link Push}input, one {@link Push}output, and one
 * {@link Push1}input.
 * <p>
 * It receives calls to the <code>push</code> method. It generates a key for
 * the message, then it pushes the message on its output. When a message with
 * the same key arrives on its <code>Push1</code> input, it triggers an
 * handler for this message.
 * <p>
 * <i>Note: </i> if a message with an "All" key is received on the input, it is
 * used as a return message of all the call to the <code>pushWithReturn</code>
 * method.
 * <p>
 * <i>Note 2: </i> if a return message does not correspond to a stored key, it
 * is forwarded.
 */
public class SynchronizerAsynchronousImpl extends AbstractComponent
    implements
      Push,
      Push1
{

  // Client interfaces
  /** The client interface used to access the message manager. */
  MessageManager      messageManagerItf;

  /** The KeyGenerator client interface. */
  KeyGenerator        keyGeneratorItf;

  /** The WaitingKeys client interface. */
  WaitingKeys         waitingKeysItf;

  /** The HandleReturnMessage client interface. */
  HandleReturnMessage handleReturnMessageItf;

  /** The Push output interface. */
  Push                outgoingOutPushItf;

  /** The Push incoming interface. */
  Push                incomingOutPushItf;

  static final Object LOCK = new Object();

  // ---------------------------------------------------------------------------
  // Implementation of the PushWithReturn interface.
  // ---------------------------------------------------------------------------

  /**
   * @see Push#push(Message)
   */
  public void push(Message message) throws PushException
  {
    Key key = null;
    try
    {
      // Generate a key for the message
      key = keyGeneratorItf.generateKey(message);
    }
    catch (Exception e)
    {
      logger.log(BasicLevel.ERROR, e);
      throw new PushException(e);
    }

    // Add the key to the HashMap
    WaitingKey waitingKey;
    synchronized (LOCK)
    {
      try
      {
        waitingKey = waitingKeysItf.createWaitingKey(key);
        // TODO Cloning policy should be set using an attribute
        messageManagerItf.duplicateMessage(message, false);
        waitingKey.setMessage(message);
      }
      catch (InterruptedException e)
      {
        throw new InterruptedPushException(e);
      }
    }

    // Push the message on the output
    outgoingOutPushItf.push(message);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Push1 interface.
  // ---------------------------------------------------------------------------

  /**
   * @see Push1#push1(Message)
   */
  public void push1(Message message) throws PushException
  {
    Key key = null;
    try
    {
      // Generate or retrieve a key for the message
      key = keyGeneratorItf.generateKey(message);
    }
    catch (Exception e)
    {
      logger.log(BasicLevel.ERROR, e);
      throw new PushException(e);
    }

    if (key != null)
    {
      if (key.isAll())
      {
        synchronized (LOCK)
        {
          Iterator iter;
          try
          {
            iter = waitingKeysItf.removeWaitingKeys(key);
          }
          catch (InterruptedException e)
          {
            throw new InterruptedPushException(e);
          }
          while (iter.hasNext())
          {
            WaitingKey waitingKeyTemp = (WaitingKey) iter.next();
            waitingKeyTemp.setReturnMessage(message);
            try
            {
              handleReturnMessageItf.handleReturnMessage(waitingKeyTemp);
            }
            catch (Exception e)
            {
              throw new PushException(e);
            }
          }
        }
      }
      else
      {
        try
        {
          synchronized (LOCK)
          {
            WaitingKey waitingKey = waitingKeysItf.removeWaitingKey(key);
            if (waitingKey != null)
            {
              waitingKey.setReturnMessage(message);
              try
              {
                handleReturnMessageItf.handleReturnMessage(waitingKey);
              }
              catch (Exception e)
              {
                throw new PushException(e);
              }
            }
            else
            {
              // Already received a key for this message
              messageManagerItf.deleteMessage(message);
            }
          }
        }
        catch (InterruptedException e)
        {
          throw new InterruptedPushException(e);
        }
      }
    }
    else
    {
      // No key --> forward the message
      incomingOutPushItf.push(message);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(Push.OUTGOING_OUT_PUSH_ITF_NAME))
    {
      outgoingOutPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(Push.INCOMING_OUT_PUSH_ITF_NAME))
    {
      incomingOutPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(KeyGenerator.ITF_NAME))
    {
      keyGeneratorItf = (KeyGenerator) serverItf;
    }
    else if (clientItfName.equals(WaitingKeys.ITF_NAME))
    {
      waitingKeysItf = (WaitingKeys) serverItf;
    }
    else if (clientItfName.equals(HandleReturnMessage.ITF_NAME))
    {
      handleReturnMessageItf = (HandleReturnMessage) serverItf;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{KeyGenerator.ITF_NAME, Push.OUT_PUSH_ITF_NAME,
        MessageManager.ITF_NAME, WaitingKeys.ITF_NAME};
  }

}
