/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.queue;

import org.objectweb.dream.message.Message;

/**
 * This interface defines methods that must be implemented by a buffer to allow
 * components to get or delete messages from it. Such an interface is used by
 * <i>consumers </i> of messages. <p/><i>Note 1: </i> all the messages that are
 * stored in a buffer are not always available for removal (e.g. when the buffer
 * implements an ordering policy). <p/><i>Note 2: </i> <b>buffer
 * implementations MUST BE thread-safe </b>.
 */
public interface BufferRemove
{

  /**
   * The commonly used name to refer to an interface with this signature.
   */
  String ITF_NAME = "buffer-remove";

  /**
   * Gets a message from the buffer. This method blocks if no message is
   * available.
   * 
   * @return a message; never <code>null</code>.
   * @throws InterruptedException if it is interrupted while waiting for a
   *           message to be added.
   */
  Message get() throws InterruptedException;

  /**
   * Gets a message from the buffer. This method does not block.
   * 
   * @return a message if there is one available, <code>null</code> otherwise.
   */
  Message tryGet();

  /**
   * Removes a message from the buffer. This method blocks if no message is
   * available.
   * 
   * @return the removed message; never <code>null</code>.
   * @throws InterruptedException if it is interrupted while waiting for a
   *           message to be added.
   */
  Message remove() throws InterruptedException;

  /**
   * Removes a message from the buffer. This method does not block.
   * 
   * @return a message if there is one available, <code>null</code> otherwise.
   */
  Message tryRemove();

  /**
   * Removes all the available messages from the buffer. Messages are returned
   * in an aggregated message. If no message has been removed, this method
   * returns <code>null</code>. This method does not block.
   * 
   * @return an aggregated message containing all the messages available from
   *         the buffer, or <code>null</code> if no message was available.
   */
  Message removeAll();

}