/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Romain Lenglet
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.queue;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.InterruptedPushException;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * An incoming handler that adds {@link #push(Message) pushed} messages into a
 * {@link BufferAdd buffer} bound to it. This component uses
 * {@link BufferAdd#add(Message) blocking calls for adding} messages into the
 * buffer.
 */
public class IncomingHandlerBlockingImpl extends AbstractComponent
    implements
      Push
{

  // ------------------------------------------------------------------------
  // Constants
  // ------------------------------------------------------------------------

  /**
   * The names of client interfaces.
   */
  protected static final String[] CLIENT_INTERFACE_NAMES = {BufferAdd.ITF_NAME};

  // ------------------------------------------------------------------------
  // Client interfaces.
  // ------------------------------------------------------------------------

  /**
   * The buffer into which pushed messages are added.
   */
  protected BufferAdd             bufferAdd;

  // ------------------------------------------------------------------------
  // Implementation of the Push interface.
  // ------------------------------------------------------------------------

  /**
   * @see Push#push(Message)
   */
  public void push(Message message) throws PushException
  {
    try
    {
      bufferAdd.add(message);
    }
    catch (InterruptedException e)
    {
      throw new InterruptedPushException(
          "interrupted while adding message into the buffer", e);
    }
  }

  // ------------------------------------------------------------------------
  // Implementation of the BindingController interface.
  // ------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return CLIENT_INTERFACE_NAMES;
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(BufferAdd.ITF_NAME))
    {
      bufferAdd = (BufferAdd) serverItf;
    }
  }

}