/**
 * DREAM
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Romain Lenglet
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.queue.keyed;

import java.util.LinkedList;
import java.util.TreeMap;

import org.objectweb.dream.message.AbstractChunk;
import org.objectweb.dream.message.Message;

/**
 * This buffer stores every incoming message according to a key that is
 * contained in the message as a {@link AbstractChunk chunk} which name can be
 * specified as a Fractal component attribute using the interface of type
 * {@link BufferAttributeControllerKeyChunkAdd}. The {@link #doRemove(Object)}
 * method removes the message with the specified key.
 */
public class BufferKeyChunkAddKeyedRemoveImpl
    extends
      AbstractBufferKeyedRemoveImpl
    implements
      BufferAttributeControllerKeyChunkAdd
{

  // ------------------------------------------------------------------------
  // Fractal component attributes.
  // ------------------------------------------------------------------------

  protected String                                      keyChunkName = "key-chunk";

  // ------------------------------------------------------------------------
  // Fields.
  // ------------------------------------------------------------------------

  /**
   * Used to sort the messages according to the key that is a chunk in every
   * message added to this buffer.
   */
  protected TreeMap<AbstractChunk, LinkedList<Message>> queue        = new TreeMap<AbstractChunk, LinkedList<Message>>();

  // ------------------------------------------------------------------------
  // Overridden methods
  // ------------------------------------------------------------------------

  @SuppressWarnings("unchecked")
  @Override
  protected void doAdd(Message message)
  {
    AbstractChunk keyChunk = messageManagerItf.getChunk(message, keyChunkName);
    if (keyChunk == null)
    {
      throw new IllegalArgumentException("message has no key chunk of name "
          + keyChunkName);
    }
    LinkedList<Message> messagesWithKey = queue.get(keyChunk);
    if (messagesWithKey == null)
    {
      messagesWithKey = new LinkedList<Message>();
      AbstractChunk clone = messageManagerItf.cloneChunk(keyChunk);
      queue.put(clone, messagesWithKey);
    }
    messagesWithKey.addLast(message);
    incrementStoredMessagesCount(1, keyChunk);
  }

  @Override
  protected Message doRemove(Object key)
  {
    LinkedList<Message> list = queue.remove(key);
    if (list != null)
    {
      incrementAvailableMessagesCount(-1, key);
      incrementStoredMessagesCount(-1, key);
    }
    return list.removeFirst();
  }

  @Override
  protected Message doGet(Object key)
  {
    LinkedList<Message> list = queue.remove(key);
    if (list != null)
    {
      return list.getFirst();
    }
    return null;
  }

  @Override
  protected Message doRemoveAll()
  {
    throw new UnsupportedOperationException();
  }

  @Override
  protected boolean canAdd(Message message)
  {
    if ((maxCapacity > 0) && (maxCapacity - storedMessagesCount <= 0))
    {
      return false;
    }
    return true;
  }

  // ------------------------------------------------------------------------
  // Implementation of the BufferAttributeControllerKeyChunkAdd interface.
  // ------------------------------------------------------------------------

  /**
   * @see BufferAttributeControllerKeyChunkAdd#setKeyChunkName(String)
   */
  public void setKeyChunkName(String keyChunkName)
  {
    this.keyChunkName = keyChunkName;
  }

  /**
   * @see BufferAttributeControllerKeyChunkAdd#getKeyChunkName()
   */
  public String getKeyChunkName()
  {
    return keyChunkName;
  }

}
