/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq, Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.synchro;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.util.EmptyStringArray;

/**
 * Basic implementation of the {@link MutexReEntrant} interface: this mutex is
 * reentrant.
 * <p>
 * <b>Note: inspired by Doug Lea's implementation. </b>
 */
public class MutexReEntrantImpl extends AbstractComponent
    implements
      MutexReEntrant
{

  /** The lock status. */
  protected boolean inuse   = false;

  ThreadLocal<Integer>       counter = new ThreadLocal<Integer>();

  // ---------------------------------------------------------------------------
  // Implementation of the MutexReEntrant interface.
  // ---------------------------------------------------------------------------

  /**
   * @see Mutex#lock()
   */
  public void lock() throws InterruptedException
  {
    if (Thread.interrupted())
      throw new InterruptedException();
    synchronized (this)
    {
      try
      {
        Integer c = counter.get();
        if (c == null)
        {
          while (inuse)
            wait();
          c = 1;
        }
        else
        {
          c = c + 1;
        }
        counter.set(c);
        inuse = true;
      }
      catch (InterruptedException ex)
      {
        // We are not re-entrant. c == null
        notify();
        throw ex;
      }
    }
  }

  /**
   * @see Mutex#timedLock(long)
   */
  public boolean timedLock(long msecs) throws InterruptedException
  {
    if (Thread.interrupted())
      throw new InterruptedException();
    synchronized (this)
    {
      Integer c = counter.get();
      if (c == null && !inuse)
      {
        inuse = true;
        counter.set(1);
        return true;
      }
      else if (c != null)
      {
        counter.set(c + 1);
        return true;
      }
      else if (msecs <= 0)
        return false;
      else
      {
        // I don't have the lock: c == null and inuse == true
        long waitTime = msecs;
        long start = System.currentTimeMillis();
        try
        {
          for (;;)
          {
            wait(waitTime);
            if (!inuse)
            {
              inuse = true;
              counter.set(1);
              return true;
            }
            waitTime = msecs - (System.currentTimeMillis() - start);
              if (waitTime <= 0)
                return false;
          }
        }
        catch (InterruptedException ex)
        {
          notify();
          throw ex;
        }
      }
    }
  }

  /**
   * @see Mutex#unlock()
   */
  public synchronized void unlock()
  {
    Integer c = counter.get();
    if (c == 1)
    {
      counter.set(null);
      inuse = false;
      notify();
    }
    else
    {
      counter.set(new Integer(c.intValue() - 1));
    }
  }

  /**
   * @see MutexReEntrant#forceUnlock()
   */
  public synchronized Object forceUnlock()
  {
    Integer c = counter.get();
    counter.set(null);
    inuse = false;
    notify();
    return c;
  }

  /**
   * @see MutexReEntrant#forceLock(Object)
   */
  public void forceLock(Object o) throws InterruptedException
  {
    if (Thread.interrupted())
      throw new InterruptedException();
    synchronized (this)
    {
      try
      {
        Integer c = counter.get();
        if (c == null)
        {
          while (inuse)
            wait();
          c = (Integer) o;
        }
        else
        {
          c = c + (Integer) o;
        }
        counter.set(c);
        inuse = true;
      }
      catch (InterruptedException ex)
      {
        // We are not re-entrant. c == null
        notify();
        throw ex;
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return EmptyStringArray.EMPTY_STRING_ARRAY;
  }

}